@@ -12,7 +12,7 @@ the full text of the license.
 '''
 
 import subprocess, tempfile, os.path, urllib, re, pwd, grp, os, sys
-import ctypes, ctypes.util, fnmatch
+import fnmatch, glob
 
 import xml.dom, xml.dom.minidom
 from xml.parsers.expat import ExpatError
@@ -22,10 +22,14 @@ import fileutils
 from packaging_impl import impl as packaging
 
 _hook_dir = '/usr/share/apport/package-hooks/'
+_common_hook_dir = '/usr/share/apport/general-hooks/'
 
 # path of the ignore file
 _ignore_file = '~/.apport-ignore.xml'
 
+# system-wide blacklist
+_blacklist_dir = '/etc/apport/blacklist.d'
+
 # programs that we consider interpreters
 interpreters = ['sh', 'bash', 'dash', 'csh', 'tcsh', 'python*',
     'ruby*', 'php', 'perl*', 'mono*', 'awk']
@@ -65,15 +69,7 @@ def _read_maps(pid):
     try:
         maps = file('/proc/%d/maps' % pid).read().strip()
     except (OSError,IOError), e:
-        try:
-            libc = ctypes.CDLL(ctypes.util.find_library("c"))
-            # PT_ATTACH
-            libc.ptrace(16, pid, 0, 0)
-            maps = _read_file('/proc/%d/maps' % pid)
-            # PT_DETACH
-            libc.ptrace(17, pid, 0, 0)
-        except (OSError, IOError), e:
-            return 'Error: ' + str(e)
+        return 'Error: ' + str(e)
     return maps
 
 def _command_output(command, input = None, stderr = subprocess.STDOUT):
@@ -125,6 +121,42 @@ def _dom_remove_space(node):
         else:
             _dom_remove_space(c)
 
+def get_module_license(module):
+    '''Return the license for a given kernel module.'''
+
+    try:
+        modinfo = subprocess.Popen(['/sbin/modinfo', module], stdout=subprocess.PIPE)
+        out = modinfo.communicate()[0]
+        if modinfo.returncode != 0:
+            return None
+    except OSError:
+        return None
+    for l in out.splitlines():
+        fields = l.split(':', 1)
+        if len(fields) < 2:
+            continue
+        if fields[0] == 'license':
+            return fields[1].strip()
+
+    return None
+
+def nonfree_modules():
+    '''Check loaded modules and return a list of those which are not free.'''
+    try:
+        mods = [l.split()[0] for l in open('/proc/modules')]
+    except IOError:
+        return []
+
+    nonfree = []
+    for m in mods:
+        l = get_module_license(m)
+        if l and ('GPL' in l or 'BSD' in l or 'MPL' in l or 'MIT' in l):
+            pass
+        else:
+            nonfree.append(m)
+
+    return nonfree
+
 #
 # Report class
 #
@@ -205,7 +237,9 @@ class Report(ProblemReport):
         This adds:
         - DistroRelease: lsb_release -sir output
         - Architecture: system architecture in distro specific notation
-        - Uname: uname -a output'''
+        - Uname: uname -a output
+        - NonfreeKernelModules: loaded kernel modules which are not free (if
+            there are none, this field will not be present)'''
 
         p = subprocess.Popen(['lsb_release', '-sir'], stdout=subprocess.PIPE,
             stderr=subprocess.STDOUT, close_fds=True)
@@ -215,6 +249,9 @@ class Report(ProblemReport):
             stderr=subprocess.STDOUT, close_fds=True)
         self['Uname'] = p.communicate()[0].strip()
         self['Architecture'] = packaging.get_system_architecture()
+        nm = nonfree_modules()
+        if nm:
+            self['NonfreeKernelModules'] = ' '.join(nonfree_modules())
 
     def add_user_info(self):
         '''Add information about the user.
@@ -241,7 +278,7 @@ class Report(ProblemReport):
         # check if we consider ExecutablePath an interpreter; we have to do
         # this, otherwise 'gedit /tmp/foo.txt' would be detected as interpreted
         # script as well
-        if not any([fnmatch.fnmatch(exebasename, i) for i in interpreters]):
+        if not filter(lambda i: fnmatch.fnmatch(exebasename, i), interpreters):
             return
 
         # first, determine process name
@@ -339,7 +376,7 @@ class Report(ProblemReport):
     def add_gdb_info(self, debugdir=None):
         '''Add information from gdb.
 
-        This requires that the report has a CoreDump (file ref) and an
+        This requires that the report has a CoreDump and an
         ExecutablePath. This adds the following fields:
         - Registers: Output of gdb's 'info registers' command
         - Disassembly: Output of gdb's 'x/16i $pc' command
@@ -362,6 +399,11 @@ class Report(ProblemReport):
                 os.write(fd, self['CoreDump'])
                 os.close(fd)
                 unlink_core = True
+            elif hasattr(self['CoreDump'], 'gzipvalue'):
+                (fd, core) = tempfile.mkstemp()
+                os.close(fd)
+                self['CoreDump'].write(open(core, 'w'))
+                unlink_core = True
             else:
                 core = self['CoreDump'][0]
 
@@ -386,9 +428,12 @@ class Report(ProblemReport):
                 command += ['--ex', 'p -99', '--ex', cmd]
 
             # call gdb
-            out = _command_output(command, stderr=open('/dev/null')).replace(
-                '(no debugging symbols found)\n','').replace(
-                'No symbol table info available.\n','')
+            try:
+                out = _command_output(command, stderr=open('/dev/null')).replace(
+                    '(no debugging symbols found)\n','').replace(
+                    'No symbol table info available.\n','')
+            except OSError:
+                return
 
             # split the output into the various fields
             part_re = re.compile('^\$\d+\s*=\s*-99$', re.MULTILINE)
@@ -401,33 +446,74 @@ class Report(ProblemReport):
             if unlink_core:
                 os.unlink(core)
 
-        # StacktraceTop
+        if self.has_key('Stacktrace'):
+            self._gen_stacktrace_top()
+
+    def _gen_stacktrace_top(self):
+        '''Build field StacktraceTop as the top five functions of Stacktrace. 
+
+        Signal handler invocations and related functions are skipped since they
+        are generally not useful for triaging and duplicate detection.'''
+        
+        unwind_functions = set(['g_logv', 'g_log', 'IA__g_log', 'IA__g_logv',
+            'g_assert_warning', 'IA__g_assert_warning'])
         toptrace = [''] * 5
-        bt_fn_re = re.compile('^#(\d+)\s+0x(?:\w+)\s+in\s+(.*)$')
+        depth = 0
+        unwound = False
+        unwinding = False
+        bt_fn_re = re.compile('^#(\d+)\s+(?:0x(?:\w+)\s+in\s+(.*)|(<signal handler called>)\s*)$')
+
         for line in self['Stacktrace'].splitlines():
             m = bt_fn_re.match(line)
-            if m:
-                depth = int(m.group(1))
-                if depth < len(toptrace):
-                    toptrace[depth] = m.group(2)
+            if not m:
+                continue
+
+            if not unwound or unwinding:
+                if m.group(2):
+                    fn = m.group(2).split()[0].split('(')[0]
+                else:
+                    fn = None
+                if m.group(3) or fn in unwind_functions:
+                    unwinding = True
+                    depth = 0
+                    toptrace = [''] * 5
+                    unwound = True
+                    continue
+                else:
+                    unwinding = False
+
+            if depth < len(toptrace):
+                toptrace[depth] = m.group(2) or m.group(3)
+                depth += 1
         self['StacktraceTop'] = '\n'.join(toptrace).strip()
 
     def add_hooks_info(self):
         '''Check for an existing hook script and run it to add additional
         package specific information.
 
-        A hook script needs to be in _hook_dir/<Package>.py and has to
-        contain a function 'add_info(report)' that takes and modifies a
-        Report.'''
+        A hook script needs to be in _hook_dir/<Package>.py or in
+        _common_hook_dir/*.py and has to contain a function 'add_info(report)'
+        that takes and modifies a Report.'''
 
         symb = {}
         assert self.has_key('Package')
+
+        # common hooks
+        for hook in glob.glob(_common_hook_dir + '/*.py'):
+            try:
+                execfile(hook, symb)
+                symb['add_info'](self)
+            except:
+                pass
+
+        # binary package hook
         try:
             execfile('%s/%s.py' % (_hook_dir, self['Package'].split()[0]), symb)
             symb['add_info'](self)
         except:
             pass
 
+        # source package hook
         if self.has_key('SourcePackage'):
             try:
                 execfile('%s/source_%s.py' % (_hook_dir, self['SourcePackage'].split()[0]), symb)
@@ -447,16 +533,20 @@ class Report(ProblemReport):
         For example:
         <?xml version="1.0"?>
         <patterns>
-            <pattern url="https://launchpad.net/bugs/1">
+            <pattern url="http://bugtracker.net/bugs/1">
                 <re key="Foo">ba.*r</re>
             </pattern>
-            <pattern url="https://launchpad.net/bugs/2">
+            <pattern url="http://bugtracker.net/bugs/2">
                 <re key="Foo">write_(hello|goodbye)</re>
                 <re key="Package">^\S* 1-2$</re> <!-- test for a particular version -->
             </pattern>
         </patterns>
         '''
 
+        # some distros might not want to support these
+        if not baseurl:
+            return
+
         assert self.has_key('Package')
         package = self['Package'].split()[0]
         try:
@@ -500,17 +590,32 @@ class Report(ProblemReport):
         return dom
 
     def check_ignored(self):
-        '''Check ~/.apport-ignore.xml (in the real UID's home) if the current
-        report should not be presented to the user.
+        '''Check ~/.apport-ignore.xml (in the real UID's home) and
+        /etc/apport/blacklist.d/ if the current report should not be presented
+        to the user.
 
         This requires the ExecutablePath attribute. Function can throw a
         ValueError if the file has an invalid format.'''
 
         assert self.has_key('ExecutablePath')
+
+        # check blacklist
+        try:
+            for f in os.listdir(_blacklist_dir):
+                try:
+                    fd = open(os.path.join(_blacklist_dir, f))
+                except IOError:
+                    continue
+                for line in fd:
+                    if line.strip() == self['ExecutablePath']:
+                        return True
+        except OSError:
+            pass
+
         dom = self._get_ignore_dom()
 
         try:
-            cur_mtime = float(os.stat(self['ExecutablePath']).st_mtime)
+            cur_mtime = int(os.stat(self['ExecutablePath']).st_mtime)
         except OSError:
             # if it does not exist any more, do nothing
             return False
@@ -554,6 +659,166 @@ class Report(ProblemReport):
 
         dom.unlink()
 
+    def has_useful_stacktrace(self):
+        '''Check whether this report has a stacktrace that can be considered
+        'useful'.
+
+        The current heuristic is to consider it useless if it either is shorter
+        than three lines and has any unknown function, or for longer traces, a
+        minority of known functions.'''
+        
+        if not self.get('StacktraceTop'):
+            return False
+        
+        unknown_fn = [f.startswith('??') for f in self['StacktraceTop'].splitlines()]
+
+        if len(unknown_fn) < 3:
+            return unknown_fn.count(True) == 0
+
+        return unknown_fn.count(True) <= len(unknown_fn)/2.
+
+    def standard_title(self):
+        '''Create an appropriate title for a crash database entry.
+
+        This contains the topmost function name from the stack trace and the
+        signal (for signal crashes) or the Python exception (for unhandled
+        Python exceptions).
+
+        Return None if the report is not a crash or a default title could not
+        be generated.'''
+
+        # signal crash
+        if self.has_key('Signal') and \
+            self.has_key('ExecutablePath') and \
+            self.has_key('StacktraceTop'):
+
+            signal_names = {
+                '4': 'SIGILL',
+                '6': 'SIGABRT',
+                '8': 'SIGFPE',
+                '11': 'SIGSEGV',
+                '13': 'SIGPIPE'
+            }
+
+            fn = ''
+            for l in self['StacktraceTop'].splitlines():
+                fname = l.split('(')[0].strip()
+                if fname != '??':
+                    fn = ' in %s()' % fname
+                    break
+
+            arch_mismatch = ''
+            if self.has_key('Architecture') and \
+                self.has_key('PackageArchitecture') and \
+                self['Architecture'] != self['PackageArchitecture'] and \
+                self['PackageArchitecture'] != 'all':
+                arch_mismatch = ' [non-native %s package]' % self['PackageArchitecture']
+
+            return '%s crashed with %s%s%s' % (
+                os.path.basename(self['ExecutablePath']),
+                signal_names.get(self.get('Signal'),
+                    'signal ' + self.get('Signal')),
+                fn, arch_mismatch
+            )
+
+        # Python exception
+        if self.has_key('Traceback') and \
+            self.has_key('ExecutablePath'):
+
+            trace = self['Traceback'].splitlines()
+
+            if len(trace) < 1:
+                return None
+            if len(trace) < 3:
+                return '%s crashed with %s' % (
+                    os.path.basename(self['ExecutablePath']),
+                    trace[0])
+            return '%s crashed with %s in %s()' % (
+                os.path.basename(self['ExecutablePath']),
+                trace[-1].split(':')[0],
+                trace[-3].split()[-1]
+            )
+
+        # package problem
+        if self.get('ProblemType') == 'Package' and \
+            self.has_key('Package'):
+
+            title = 'package %s failed to install/upgrade' % \
+                self['Package']
+            if self.get('ErrorMessage'):
+                title += ': ' + self['ErrorMessage'].splitlines()[-1]
+
+            return title
+
+        return None
+
+    def obsolete_packages(self):
+        '''Check Package: and Dependencies: for obsolete packages and return a
+        list of them.'''
+
+        obsolete = []
+        for l in (self['Package'] + '\n' + self.get('Dependencies', '')).splitlines():
+            if not l:
+                continue
+            pkg, ver = l.split()[:2]
+            avail = packaging.get_available_version(pkg)
+            if ver != None and ver != 'None' and avail != None and \
+                packaging.compare_versions(ver, avail) < 0:
+                obsolete.append(pkg)
+        return obsolete
+
+    def crash_signature(self):
+        '''Calculate a signature string for a crash suitable for identifying
+        duplicates.
+
+        For signal crashes this the concatenation of ExecutablePath, Signal
+        number, and StacktraceTop function names, separated by a colon. If
+        StacktraceTop has unknown functions or the report lacks any of those
+        fields, return None.
+        
+        For Python crashes, this concatenates the ExecutablePath, exception
+        name, and Traceback function names, again separated by a colon.'''
+
+        if not self.has_key('ExecutablePath'):
+            return None
+
+        # signal crashes
+        if self.has_key('StacktraceTop') and self.has_key('Signal'):
+            sig = '%s:%s' % (self['ExecutablePath'], self['Signal'])
+            bt_fn_re = re.compile('^(?:([\w:~]+).*|(<signal handler called>)\s*)$')
+
+            for line in self['StacktraceTop'].splitlines():
+                m = bt_fn_re.match(line)
+                if m:
+                    sig += ':' + (m.group(1) or m.group(2))
+                else:
+                    # this will also catch ??
+                    return None
+            return sig
+
+        # Python crashes
+        if self.has_key('Traceback'):
+            trace = self['Traceback'].splitlines()
+
+            sig = ''
+            if len(trace) == 1:
+                # sometimes, Python exceptions do not have file references
+                m = re.match('(\w+): ', trace[0])
+                if m:
+                    return self['ExecutablePath'] + ':' + m.group(1)
+                else:
+                    return None
+            elif len(trace) < 3:
+                return None
+
+            for l in trace:
+                if l.startswith('  File'):
+                    sig += ':' + l.split()[-1]
+
+            return self['ExecutablePath'] + ':' + trace[-1].split(':')[0] + sig
+
+        return None
+
 #
 # Unit test
 #
@@ -687,7 +952,8 @@ class _ApportReportTest(unittest.TestCase):
         pr.add_proc_info(pid=p.pid)
         p.communicate('\n')
         self.assertEqual(pr['ExecutablePath'], '/bin/zgrep')
-        self.assertEqual(pr['InterpreterPath'], os.path.realpath('/bin/sh'))
+        self.assertEqual(pr['InterpreterPath'],
+            os.path.realpath(open('/bin/zgrep').readline().strip()[2:]))
         self.assertTrue('[stack]' in pr['ProcMaps'])
 
         # check correct handling of interpreted executables: python
@@ -968,10 +1234,10 @@ gdb --batch --ex 'generate-core-file %s' --pid $$ >/dev/null''' % coredump)
             # create some test patterns
             open(os.path.join(pdir, 'bash.xml'), 'w').write('''<?xml version="1.0"?>
 <patterns>
-    <pattern url="https://launchpad.net/bugs/1">
+    <pattern url="http://bugtracker.net/bugs/1">
         <re key="Foo">ba.*r</re>
     </pattern>
-    <pattern url="https://launchpad.net/bugs/2">
+    <pattern url="http://bugtracker.net/bugs/2">
         <re key="Foo">write_(hello|goodbye)</re>
         <re key="Package">^\S* 1-2$</re>
     </pattern>
@@ -979,10 +1245,10 @@ gdb --batch --ex 'generate-core-file %s' --pid $$ >/dev/null''' % coredump)
 
             open(os.path.join(pdir, 'coreutils.xml'), 'w').write('''<?xml version="1.0"?>
 <patterns>
-    <pattern url="https://launchpad.net/bugs/3">
+    <pattern url="http://bugtracker.net/bugs/3">
         <re key="Bar">^1$</re>
     </pattern>
-    <pattern url="https://launchpad.net/bugs/4">
+    <pattern url="http://bugtracker.net/bugs/4">
         <re key="Bar">*</re> <!-- invalid RE -->
     </pattern>
 </patterns>''')
@@ -1004,10 +1270,10 @@ gdb --batch --ex 'generate-core-file %s' --pid $$ >/dev/null''' % coredump)
             r_invalid['Package'] = 'invalid 1'
 
             # positive match cases
-            self.assertEqual(r_bash.search_bug_patterns(pdir), 'https://launchpad.net/bugs/1')
+            self.assertEqual(r_bash.search_bug_patterns(pdir), 'http://bugtracker.net/bugs/1')
             r_bash['Foo'] = 'write_goodbye'
-            self.assertEqual(r_bash.search_bug_patterns(pdir), 'https://launchpad.net/bugs/2')
-            self.assertEqual(r_coreutils.search_bug_patterns(pdir), 'https://launchpad.net/bugs/3')
+            self.assertEqual(r_bash.search_bug_patterns(pdir), 'http://bugtracker.net/bugs/2')
+            self.assertEqual(r_coreutils.search_bug_patterns(pdir), 'http://bugtracker.net/bugs/3')
 
             # negative match cases
             r_bash['Package'] = 'bash 1-21'
@@ -1032,8 +1298,10 @@ gdb --batch --ex 'generate-core-file %s' --pid $$ >/dev/null''' % coredump)
                 'gracefully handles nonexisting package XML file')
             self.assertEqual(r_bash.search_bug_patterns('file:///nonexisting/directory/'), None,
                 'gracefully handles nonexisting base path')
+            # existing host, but no bug patterns
             self.assertEqual(r_bash.search_bug_patterns('http://security.ubuntu.com/'), None,
                 'gracefully handles base path without bug patterns')
+            # nonexisting host
             self.assertEqual(r_bash.search_bug_patterns('http://nonexisting.domain/'), None,
                 'gracefully handles nonexisting URL domain')
         finally:
@@ -1044,14 +1312,32 @@ gdb --batch --ex 'generate-core-file %s' --pid $$ >/dev/null''' % coredump)
         '''Test add_hooks_info().'''
 
         global _hook_dir
+        global _common_hook_dir
         orig_hook_dir = _hook_dir
         _hook_dir = tempfile.mkdtemp()
+        orig_common_hook_dir = _common_hook_dir
+        _common_hook_dir = tempfile.mkdtemp()
         try:
             open(os.path.join(_hook_dir, 'foo.py'), 'w').write('''
 def add_info(report):
     report['Field1'] = 'Field 1'
     report['Field2'] = 'Field 2\\nBla'
 ''')
+
+            open(os.path.join(_common_hook_dir, 'foo1.py'), 'w').write('''
+def add_info(report):
+    report['CommonField1'] = 'CommonField 1'
+''')
+            open(os.path.join(_common_hook_dir, 'foo2.py'), 'w').write('''
+def add_info(report):
+    report['CommonField2'] = 'CommonField 2'
+''')
+
+            # should only catch .py files
+            open(os.path.join(_common_hook_dir, 'notme'), 'w').write('''
+def add_info(report):
+    report['BadField'] = 'XXX'
+''')
             r = Report()
             self.assertRaises(AssertionError, r.add_hooks_info)
 
@@ -1060,30 +1346,38 @@ def add_info(report):
             # should not throw any exceptions
             r.add_hooks_info()
             self.assertEqual(set(r.keys()), set(['ProblemType', 'Date',
-                'Package']), 'report has required fields')
+                'Package', 'CommonField1', 'CommonField2']), 
+                'report has required fields')
 
             r = Report()
             r['Package'] = 'baz 1.2-3'
             # should not throw any exceptions
             r.add_hooks_info()
             self.assertEqual(set(r.keys()), set(['ProblemType', 'Date',
-                'Package']), 'report has required fields')
+                'Package', 'CommonField1', 'CommonField2']), 
+                'report has required fields')
 
             r = Report()
             r['Package'] = 'foo'
             r.add_hooks_info()
             self.assertEqual(set(r.keys()), set(['ProblemType', 'Date',
-                'Package', 'Field1', 'Field2']), 'report has required fields')
+                'Package', 'Field1', 'Field2', 'CommonField1',
+                'CommonField2']), 'report has required fields')
             self.assertEqual(r['Field1'], 'Field 1')
             self.assertEqual(r['Field2'], 'Field 2\nBla')
+            self.assertEqual(r['CommonField1'], 'CommonField 1')
+            self.assertEqual(r['CommonField2'], 'CommonField 2')
 
             r = Report()
             r['Package'] = 'foo 4.5-6'
             r.add_hooks_info()
             self.assertEqual(set(r.keys()), set(['ProblemType', 'Date',
-                'Package', 'Field1', 'Field2']), 'report has required fields')
+                'Package', 'Field1', 'Field2', 'CommonField1',
+                'CommonField2']), 'report has required fields')
             self.assertEqual(r['Field1'], 'Field 1')
             self.assertEqual(r['Field2'], 'Field 2\nBla')
+            self.assertEqual(r['CommonField1'], 'CommonField 1')
+            self.assertEqual(r['CommonField2'], 'CommonField 2')
 
             # source package hook
             open(os.path.join(_hook_dir, 'source_foo.py'), 'w').write('''
@@ -1096,14 +1390,18 @@ def add_info(report):
             r['Package'] = 'libfoo 3'
             r.add_hooks_info()
             self.assertEqual(set(r.keys()), set(['ProblemType', 'Date',
-                'Package', 'SourcePackage', 'Field1', 'Field2']), 
-                'report has required fields')
+                'Package', 'SourcePackage', 'Field1', 'Field2', 'CommonField1',
+                'CommonField2']), 'report has required fields')
             self.assertEqual(r['Field1'], 'Field 1')
             self.assertEqual(r['Field2'], 'Field 2\nBla')
+            self.assertEqual(r['CommonField1'], 'CommonField 1')
+            self.assertEqual(r['CommonField2'], 'CommonField 2')
 
         finally:
             shutil.rmtree(_hook_dir)
+            shutil.rmtree(_common_hook_dir)
             _hook_dir = orig_hook_dir
+            _common_hook_dir = orig_common_hook_dir
 
     def test_ignoring(self):
         '''Test mark_ignore() and check_ignored().'''
@@ -1157,5 +1455,341 @@ def add_info(report):
             shutil.rmtree(workdir)
             _ignore_file = orig_ignore_file
 
+    def test_blacklisting(self):
+        '''Test check_ignored() for system-wise blacklist.'''
+
+        global _blacklist_dir
+        global _ignore_file
+        orig_blacklist_dir = _blacklist_dir
+        _blacklist_dir = tempfile.mkdtemp()
+        orig_ignore_file = _ignore_file
+        _ignore_file = '/nonexistant'
+        try:
+            bash_rep = Report()
+            bash_rep['ExecutablePath'] = '/bin/bash'
+            crap_rep = Report()
+            crap_rep['ExecutablePath'] = '/bin/crap'
+
+            # no ignores initially
+            self.assertEqual(bash_rep.check_ignored(), False)
+            self.assertEqual(crap_rep.check_ignored(), False)
+
+            # should not stumble over comments
+            open(os.path.join(_blacklist_dir, 'README'), 'w').write(
+                '# Ignore file\n#/bin/bash\n')
+
+            # no ignores on nonmatching paths
+            open(os.path.join(_blacklist_dir, 'bl1'), 'w').write(
+                '/bin/bas\n/bin/bashh\nbash\nbin/bash\n')
+            self.assertEqual(bash_rep.check_ignored(), False)
+            self.assertEqual(crap_rep.check_ignored(), False)
+
+            # ignore crap now
+            open(os.path.join(_blacklist_dir, 'bl_2'), 'w').write(
+                '/bin/crap\n')
+            self.assertEqual(bash_rep.check_ignored(), False)
+            self.assertEqual(crap_rep.check_ignored(), True)
+
+            # ignore bash now
+            open(os.path.join(_blacklist_dir, 'bl1'), 'a').write(
+                '/bin/bash\n')
+            self.assertEqual(bash_rep.check_ignored(), True)
+            self.assertEqual(crap_rep.check_ignored(), True)
+        finally:
+            shutil.rmtree(_blacklist_dir)
+            _blacklist_dir = orig_blacklist_dir
+            _ignore_file = orig_ignore_file
+
+    def test_has_useful_stacktrace(self):
+        '''Test has_useful_stacktrace().'''
+
+        r = Report()
+        self.failIf(r.has_useful_stacktrace())
+
+        r['StacktraceTop'] = ''
+        self.failIf(r.has_useful_stacktrace())
+
+        r['StacktraceTop'] = '?? ()'
+        self.failIf(r.has_useful_stacktrace())
+
+        r['StacktraceTop'] = '?? ()\n?? ()'
+        self.failIf(r.has_useful_stacktrace())
+
+        r['StacktraceTop'] = 'read () from /lib/libc.6.so\n?? ()'
+        self.failIf(r.has_useful_stacktrace())
+
+        r['StacktraceTop'] = 'read () from /lib/libc.6.so\n?? ()\n?? ()\n?? ()'
+        self.failIf(r.has_useful_stacktrace())
+
+        r['StacktraceTop'] = 'read () from /lib/libc.6.so\nfoo (i=1) from /usr/lib/libfoo.so'
+        self.assert_(r.has_useful_stacktrace())
+
+        r['StacktraceTop'] = 'read () from /lib/libc.6.so\nfoo (i=1) from /usr/lib/libfoo.so\n?? ()'
+        self.assert_(r.has_useful_stacktrace())
+
+        r['StacktraceTop'] = 'read () from /lib/libc.6.so\nfoo (i=1) from /usr/lib/libfoo.so\n?? ()\n?? ()'
+        self.assert_(r.has_useful_stacktrace())
+
+        r['StacktraceTop'] = 'read () from /lib/libc.6.so\n?? ()\nfoo (i=1) from /usr/lib/libfoo.so\n?? ()\n?? ()'
+        self.failIf(r.has_useful_stacktrace())
+
+    def test_standard_title(self):
+        '''Test standard_title().'''
+
+        report = Report()
+        self.assertEqual(report.standard_title(), None)
+
+        # named signal crash
+        report['Signal'] = '11'
+        report['ExecutablePath'] = '/bin/bash'
+        report['StacktraceTop'] = '''foo()
+bar(x=3)
+baz()
+'''
+        self.assertEqual(report.standard_title(),
+            'bash crashed with SIGSEGV in foo()')
+
+        # unnamed signal crash
+        report['Signal'] = '42'
+        self.assertEqual(report.standard_title(),
+            'bash crashed with signal 42 in foo()')
+
+        # do not crash on empty StacktraceTop
+        report['StacktraceTop'] = ''
+        self.assertEqual(report.standard_title(),
+            'bash crashed with signal 42')
+
+        # do not create bug title with unknown function name
+        report['StacktraceTop'] = '??()\nfoo()'
+        self.assertEqual(report.standard_title(),
+            'bash crashed with signal 42 in foo()')
+
+        # if we do not know any function name, don't mention ??
+        report['StacktraceTop'] = '??()\n??()'
+        self.assertEqual(report.standard_title(),
+            'bash crashed with signal 42')
+
+        # Python crash
+        report = Report()
+        report['ExecutablePath'] = '/usr/share/apport/apport-gtk'
+        report['Traceback'] = '''Traceback (most recent call last):
+File "/usr/share/apport/apport-gtk", line 202, in <module>
+app.run_argv()
+File "/var/lib/python-support/python2.5/apport/ui.py", line 161, in run_argv
+self.run_crashes()
+File "/var/lib/python-support/python2.5/apport/ui.py", line 104, in run_crashes
+self.run_crash(f)
+File "/var/lib/python-support/python2.5/apport/ui.py", line 115, in run_crash
+response = self.ui_present_crash(desktop_entry)
+File "/usr/share/apport/apport-gtk", line 67, in ui_present_crash
+subprocess.call(['pgrep', '-x',
+NameError: global name 'subprocess' is not defined'''
+        self.assertEqual(report.standard_title(),
+            'apport-gtk crashed with NameError in ui_present_crash()')
+
+        # slightly weird Python crash
+        report = Report()
+        report['ExecutablePath'] = '/usr/share/apport/apport-gtk'
+        report['Traceback'] = '''TypeError: Cannot create a consistent method resolution
+order (MRO) for bases GObject, CanvasGroupableIface, CanvasGroupable'''
+        self.assertEqual(report.standard_title(),
+            'apport-gtk crashed with TypeError: Cannot create a consistent method resolution')
+
+        # package install problem
+        report = Report('Package')
+        report['Package'] = 'bash'
+
+        # no ErrorMessage
+        self.assertEqual(report.standard_title(),
+            'package bash failed to install/upgrade')
+
+        # empty ErrorMessage
+        report['ErrorMessage'] = ''
+        self.assertEqual(report.standard_title(),
+            'package bash failed to install/upgrade')
+
+        # nonempty ErrorMessage
+        report['ErrorMessage'] = 'botched\nnot found\n'
+        self.assertEqual(report.standard_title(),
+            'package bash failed to install/upgrade: not found')
+
+        # matching package/system architectures
+        report['Signal'] = '11'
+        report['ExecutablePath'] = '/bin/bash'
+        report['StacktraceTop'] = '''foo()
+bar(x=3)
+baz()
+'''
+        report['PackageArchitecture'] = 'amd64'
+        report['Architecture'] = 'amd64'
+        self.assertEqual(report.standard_title(),
+            'bash crashed with SIGSEGV in foo()')
+
+        # non-native package (on multiarch)
+        report['PackageArchitecture'] = 'i386'
+        self.assertEqual(report.standard_title(),
+            'bash crashed with SIGSEGV in foo() [non-native i386 package]')
+
+        # Arch: all package (matches every system architecture)
+        report['PackageArchitecture'] = 'all'
+        self.assertEqual(report.standard_title(),
+            'bash crashed with SIGSEGV in foo()')
+
+    def test_obsolete_packages(self):
+        '''Test obsolete_packages().'''
+
+        report = Report()
+        self.assertRaises(KeyError, report.obsolete_packages)
+
+        # should work without Dependencies
+        report['Package'] = 'bash 0'
+        self.assertEqual(report.obsolete_packages(), ['bash'])
+        report['Package'] = 'bash 0 [modified: /bin/bash]'
+        self.assertEqual(report.obsolete_packages(), ['bash'])
+        report['Package'] = 'bash ' + packaging.get_available_version('bash')
+        self.assertEqual(report.obsolete_packages(), [])
+
+        report['Dependencies'] = 'coreutils 0\ncron 0\n'
+        self.assertEqual(report.obsolete_packages(), ['coreutils', 'cron'])
+
+        report['Dependencies'] = 'coreutils %s [modified: /bin/mount]\ncron 0\n' % \
+            packaging.get_available_version('coreutils')
+        self.assertEqual(report.obsolete_packages(), ['cron'])
+
+        report['Dependencies'] = 'coreutils %s\ncron %s\n' % (
+            packaging.get_available_version('coreutils'),
+            packaging.get_available_version('cron'))
+        self.assertEqual(report.obsolete_packages(), [])
+
+    def test_gen_stacktrace_top(self):
+        '''Test _gen_stacktrace_top().'''
+        
+        # nothing to chop off
+        r = Report()
+        r['Stacktrace'] = '''#0  0x10000488 in h (p=0x0) at crash.c:25
+#1  0x100004c8 in g (x=1, y=42) at crash.c:26
+#2  0x10000514 in f (x=1) at crash.c:27
+#3  0x10000530 in e (x=1) at crash.c:28
+#4  0x10000530 in d (x=1) at crash.c:29
+#5  0x10000530 in c (x=1) at crash.c:30
+#6  0x10000550 in main () at crash.c:31
+'''
+        r._gen_stacktrace_top()
+        self.assertEqual(r['StacktraceTop'], '''h (p=0x0) at crash.c:25
+g (x=1, y=42) at crash.c:26
+f (x=1) at crash.c:27
+e (x=1) at crash.c:28
+d (x=1) at crash.c:29''')
+
+        # single signal handler invocation
+        r = Report()
+        r['Stacktrace'] = '''#0  0x10000488 in raise () from /lib/libpthread.so.0
+#1  0x100004c8 in ??
+#2  <signal handler called>
+#3  0x10000530 in e (x=1) at crash.c:28
+#4  0x10000530 in d (x=1) at crash.c:29
+#5  0x10000530 in c (x=1) at crash.c:30
+#6  0x10000550 in main () at crash.c:31
+'''
+        r._gen_stacktrace_top()
+        self.assertEqual(r['StacktraceTop'], '''e (x=1) at crash.c:28
+d (x=1) at crash.c:29
+c (x=1) at crash.c:30
+main () at crash.c:31''')
+
+        # stacked signal handler; should only cut the first one
+        r = Report()
+        r['Stacktrace'] = '''#0  0x10000488 in raise () from /lib/libpthread.so.0
+#1  0x100004c8 in ??
+#2  <signal handler called>
+#3  0x10000530 in e (x=1) at crash.c:28
+#4  0x10000530 in d (x=1) at crash.c:29
+#5  0x10000123 in raise () from /lib/libpthread.so.0
+#6  <signal handler called>
+#7  0x10000530 in c (x=1) at crash.c:30
+#8  0x10000550 in main () at crash.c:31
+'''
+        r._gen_stacktrace_top()
+        self.assertEqual(r['StacktraceTop'], '''e (x=1) at crash.c:28
+d (x=1) at crash.c:29
+raise () from /lib/libpthread.so.0
+<signal handler called>
+c (x=1) at crash.c:30''')
+
+        # Gnome assertion; should unwind the logs and assert call
+        r = Report()
+        r['Stacktrace'] = '''#0  0xb7d39cab in IA__g_logv (log_domain=<value optimized out>, log_level=G_LOG_LEVEL_ERROR, 
+    format=0xb7d825f0 "file %s: line %d (%s): assertion failed: (%s)", args1=0xbfee8e3c "xxx") at /build/buildd/glib2.0-2.13.5/glib/gmessages.c:493
+#1  0xb7d39f29 in IA__g_log (log_domain=0xb7edbfd0 "libgnomevfs", log_level=G_LOG_LEVEL_ERROR, 
+    format=0xb7d825f0 "file %s: line %d (%s): assertion failed: (%s)") at /build/buildd/glib2.0-2.13.5/glib/gmessages.c:517
+#2  0xb7d39fa6 in IA__g_assert_warning (log_domain=0xb7edbfd0 "libgnomevfs", file=0xb7ee1a26 "gnome-vfs-volume.c", line=254, 
+    pretty_function=0xb7ee1920 "gnome_vfs_volume_unset_drive_private", expression=0xb7ee1a39 "volume->priv->drive == drive")
+    at /build/buildd/glib2.0-2.13.5/glib/gmessages.c:552
+No locals.
+#3  0xb7ec6c11 in gnome_vfs_volume_unset_drive_private (volume=0x8081a30, drive=0x8078f00) at gnome-vfs-volume.c:254
+        __PRETTY_FUNCTION__ = "gnome_vfs_volume_unset_drive_private"
+#4  0x08054db8 in _gnome_vfs_volume_monitor_disconnected (volume_monitor=0x8070400, drive=0x8078f00) at gnome-vfs-volume-monitor.c:963
+        vol_list = (GList *) 0x8096d30
+        current_vol = (GList *) 0x8097470
+#5  0x0805951e in _hal_device_removed (hal_ctx=0x8074da8, udi=0x8093be4 "/org/freedesktop/Hal/devices/volume_uuid_92FC9DFBFC9DDA35")
+    at gnome-vfs-hal-mounts.c:1316
+        backing_udi = <value optimized out>
+#6  0xb7ef1ead in filter_func (connection=0x8075288, message=0x80768d8, user_data=0x8074da8) at libhal.c:820
+        udi = <value optimized out>
+        object_path = 0x8076d40 "/org/freedesktop/Hal/Manager"
+        error = {name = 0x0, message = 0x0, dummy1 = 1, dummy2 = 0, dummy3 = 0, dummy4 = 1, dummy5 = 0, padding1 = 0xb7e50c00}
+#7  0xb7e071d2 in dbus_connection_dispatch (connection=0x8075288) at dbus-connection.c:4267
+#8  0xb7e33dfd in ?? () from /usr/lib/libdbus-glib-1.so.2'''
+        r._gen_stacktrace_top()
+        self.assertEqual(r['StacktraceTop'], '''gnome_vfs_volume_unset_drive_private (volume=0x8081a30, drive=0x8078f00) at gnome-vfs-volume.c:254
+_gnome_vfs_volume_monitor_disconnected (volume_monitor=0x8070400, drive=0x8078f00) at gnome-vfs-volume-monitor.c:963
+_hal_device_removed (hal_ctx=0x8074da8, udi=0x8093be4 "/org/freedesktop/Hal/devices/volume_uuid_92FC9DFBFC9DDA35")
+filter_func (connection=0x8075288, message=0x80768d8, user_data=0x8074da8) at libhal.c:820
+dbus_connection_dispatch (connection=0x8075288) at dbus-connection.c:4267''')
+
+    def test_crash_signature(self):
+        '''Test crash_signature().'''
+
+        r = Report()
+        self.assertEqual(r.crash_signature(), None)
+
+        # signal crashes
+        r['Signal'] = '42'
+        r['ExecutablePath'] = '/bin/crash'
+
+        r['StacktraceTop'] = '''foo_bar (x=1) at crash.c:28
+d01 (x=1) at crash.c:29
+raise () from /lib/libpthread.so.0
+<signal handler called>
+__frob::~frob (x=1) at crash.c:30'''
+
+        self.assertEqual(r.crash_signature(), '/bin/crash:42:foo_bar:d01:raise:<signal handler called>:__frob::~frob')
+
+        r['StacktraceTop'] = '''foo_bar (x=1) at crash.c:28
+??
+raise () from /lib/libpthread.so.0
+<signal handler called>
+__frob (x=1) at crash.c:30'''
+        self.assertEqual(r.crash_signature(), None)
+
+        # Python crashes
+        del r['Signal']
+        r['Traceback'] = '''Traceback (most recent call last):
+  File "test.py", line 7, in <module>
+    print _f(5)
+  File "test.py", line 5, in _f
+    return g_foo00(x+1)
+  File "test.py", line 2, in g_foo00
+    return x/0
+ZeroDivisionError: integer division or modulo by zero'''
+        self.assertEqual(r.crash_signature(), '/bin/crash:ZeroDivisionError:<module>:_f:g_foo00')
+
+        # sometimes Python traces do not have file references
+        r['Traceback'] = 'TypeError: function takes exactly 0 arguments (1 given)'
+        self.assertEqual(r.crash_signature(), '/bin/crash:TypeError')
+
+        r['Traceback'] = 'FooBar'
+        self.assertEqual(r.crash_signature(), None)
+
 if __name__ == '__main__':
     unittest.main()
