@@ -14,15 +14,13 @@ the full text of the license.
 '''
 
 import glob, sys, os.path, optparse, time, traceback, locale, gettext
-import tempfile, pwd, errno, urllib
+import pwd, errno, urllib, zlib
 import subprocess, threading, webbrowser, xdg.DesktopEntry
 from gettext import gettext as _
 
-import launchpadBugs.storeblob
-
 import apport, apport.fileutils, REThread
 
-bugpattern_baseurl = 'http://people.ubuntu.com/~pitti/bugpatterns'
+from apport.crashdb import get_crashdb
 
 def thread_collect_info(report, reportfile, package):
     '''Encapsulate call to add_*_info() and update given report,
@@ -40,23 +38,24 @@ def thread_collect_info(report, reportfile, package):
     report.add_hooks_info()
     report.add_os_info()
 
-    # determine package origin
-    try:
-        this_os = report['DistroRelease'].split()[0]
-        import warnings
-        warnings.filterwarnings('ignore', 'apt API not stable yet', FutureWarning)
-        import apt
-        os_origin = False
-        origins = apt.Cache()[report['Package'].split()[0]].candidateOrigin
-        if origins:
-            for o in origins:
-                if o.origin == this_os:
-                    os_origin = True
-                    break
-        if not os_origin:
-            report['UnreportableReason'] = _('This is not a genuine %s package') % this_os
-    except (ImportError, KeyError, SystemError):
-        pass
+    # add title
+    title = report.standard_title()
+    if title:
+        report['Title'] = title
+
+    # check package origin
+    if not apport.packaging.is_distro_package(report['Package'].split()[0]):
+        #TRANS: %s is the name of the operating system
+        report['UnreportableReason'] = _('This is not a genuine %s package') % \
+            report['DistroRelease'].split()[0]
+
+    # check obsolete packages
+    if report['ProblemType'] == 'Crash':
+        old_pkgs = report.obsolete_packages()
+        if old_pkgs:
+            report['UnreportableReason'] = _('You have some obsolete package \
+versions installed. Please upgrade the following packages and check if the \
+problem still occurs:\n\n%s') % ', '.join(old_pkgs)
 
     if reportfile:
         f = open(reportfile, 'a')
@@ -66,27 +65,6 @@ def thread_collect_info(report, reportfile, package):
         apport.fileutils.mark_report_seen(reportfile)
         os.chmod (reportfile, 0600)
 
-def upload_launchpad_blob(report):
-    '''Upload given problem report to Malone and return the ticket for it.
-
-    Return None on error.'''
-
-    ticket = None
-
-    # set retracing tag
-    preamble = None
-    if report.has_key('CoreDump') and report.has_key('PackageArchitecture'):
-        a = report['PackageArchitecture']
-        preamble = 'Tags: need-%s-retrace' % a
-
-    # write MIME/Multipart version into temporary file
-    mime = tempfile.TemporaryFile()
-    report.write_mime(mime, preamble=preamble)
-    mime.flush()
-    mime.seek(0)
-
-    return launchpadBugs.storeblob.upload(mime)
-
 class UserInterface:
     '''Abstract base class for encapsulating the workflow and common code for
        any user interface implementation (like GTK, Qt, or CLI).
@@ -101,6 +79,13 @@ class UserInterface:
         self.report_file = None
         self.cur_package = None
 
+        try:
+            self.crashdb = get_crashdb(None)
+        except ImportError, e:
+            # this can happen while upgrading python packages
+            print >> sys.stderr, 'Could not import module, is a package upgrade in progress? Error:', e
+            sys.exit(1)
+
         gettext.textdomain(self.gettext_domain)
         self.parse_argv()
 
@@ -115,9 +100,12 @@ class UserInterface:
         for f in apport.fileutils.get_new_reports():
             self.run_crash(f)
 
-    def run_crash(self, report_file):
+    def run_crash(self, report_file, confirm=True):
         '''Present given crash report to the user, ask him what to do about it,
-        and offer to file a bug for it.'''
+        and offer to file a bug for it.
+        
+        If confirm is False, the user will not be asked whether to report the
+        problem.'''
 
         self.report_file = report_file
 
@@ -152,7 +140,9 @@ free memory to automatically analyze the problem and send a report to the develo
                 return
 
             # ask the user about what to do with the current crash
-            if self.report.get('ProblemType') == 'Package':
+            if not confirm:
+                pass
+            elif self.report.get('ProblemType') == 'Package':
                 response = self.ui_present_package_error()
                 if response == 'cancel':
                     return
@@ -235,11 +225,24 @@ free memory to automatically analyze the problem and send a report to the develo
             # application prematurely
             if e.errno == errno.ENOENT:
                 return
+            elif e.errno == errno.EACCES:
+                self.ui_error_message(_('Permission denied'), 
+                    _('The specified process does not belong to you. Please run this program as the process owner or as root.'))
+                return
             else:
                 raise
         self.cur_package = self.options.package
 
-        self.collect_info()
+        try:
+            self.collect_info()
+        except ValueError, e:
+            if e.message == 'package does not exist':
+                self.ui_error_message(_('Invalid problem report'), 
+                    _('Package %s does not exist') % self.cur_package)
+                return
+            else:
+                raise
+
         if not self.handle_duplicate():
             self.file_report()
 
@@ -248,6 +251,11 @@ free memory to automatically analyze the problem and send a report to the develo
 
         if self.options.filebug:
             self.run_report_bug()
+        elif self.options.crash_file:
+            try:
+                self.run_crash(self.options.crash_file, False)
+            except OSError, e:
+                self.ui_error_message(_('Invalid problem report'), str(e))
         else:
             self.run_crashes()
 
@@ -269,6 +277,9 @@ free memory to automatically analyze the problem and send a report to the develo
         optparser.add_option('-P', '--pid',
             help='Specify a running program in --file-bug mode. If this is specified, the bug report will contain more information.',
             action='store', type='string', dest='pid', default=None)
+        optparser.add_option('-c', '--crash-file',
+            help='Report the crash from given .crash file instead of the pending ones in ' + apport.fileutils.report_dir,
+            action='store', type='string', dest='crash_file', default=None, metavar='PATH')
 
         (self.options, self.args) = optparser.parse_args()
 
@@ -323,7 +334,7 @@ free memory to automatically analyze the problem and send a report to the develo
             # display a progress dialog
             self.ui_start_info_collection_progress()
 
-            if self.report['ProblemType'] != 'Kernel' and not self.report.has_key('Package'):
+            if self.report['ProblemType'] != 'Kernel' and not self.report.has_key('Stacktrace'):
                 icthread = REThread.REThread(target=thread_collect_info,
                     name='thread_collect_info',
                     args=(self.report, self.report_file, self.cur_package))
@@ -338,7 +349,7 @@ free memory to automatically analyze the problem and send a report to the develo
 
             if self.report['ProblemType'] == 'Kernel' or self.report.has_key('Package'):
                 bpthread = REThread.REThread(target=self.report.search_bug_patterns,
-                    args=(bugpattern_baseurl,))
+                    args=(self.crashdb.get_bugpattern_baseurl(),))
                 bpthread.start()
                 while bpthread.isAlive():
                     self.ui_pulse_info_collection_progress()
@@ -361,81 +372,6 @@ free memory to automatically analyze the problem and send a report to the develo
                 self.ui_shutdown()
                 sys.exit(1)
 
-    def create_crash_bug_title(self):
-        '''Create an appropriate standard bug title for a crash report.
-
-        This contains the topmost function name from the stack trace and the
-        signal (for signal crashes) or the Python exception (for unhandled
-        Python exceptions).
-
-        Return None if the report is not a crash or a default title could not
-        be generated.'''
-
-        # signal crash
-        if self.report.has_key('Signal') and \
-            self.report.has_key('ExecutablePath') and \
-            self.report.has_key('StacktraceTop'):
-
-            signal_names = {
-                '4': 'SIGILL',
-                '6': 'SIGABRT',
-                '8': 'SIGFPE',
-                '11': 'SIGSEGV',
-                '13': 'SIGPIPE'
-            }
-
-            fn = ''
-            for l in self.report['StacktraceTop'].splitlines():
-                fname = l.split('(')[0].strip()
-                if fname != '??':
-                    fn = ' in %s()' % fname
-                    break
-
-            arch_mismatch = ''
-            if self.report.has_key('Architecture') and \
-                self.report.has_key('PackageArchitecture') and \
-                self.report['Architecture'] != self.report['PackageArchitecture'] and \
-                self.report['PackageArchitecture'] != 'all':
-                arch_mismatch = ' [non-native %s package]' % self.report['PackageArchitecture']
-
-            return '[apport] %s crashed with %s%s%s' % (
-                os.path.basename(self.report['ExecutablePath']),
-                signal_names.get(self.report.get('Signal'),
-                    'signal ' + self.report.get('Signal')),
-                fn, arch_mismatch
-            )
-
-        # Python exception
-        if self.report.has_key('Traceback') and \
-            self.report.has_key('ExecutablePath'):
-
-            trace = self.report['Traceback'].splitlines()
-
-            if len(trace) < 1:
-                return Nonr
-            if len(trace) < 3:
-                return '[apport] %s crashed with %s' % (
-                    os.path.basename(self.report['ExecutablePath']),
-                    trace[0])
-            return '[apport] %s crashed with %s in %s()' % (
-                os.path.basename(self.report['ExecutablePath']),
-                trace[-1].split(':')[0],
-                trace[-3].split()[-1]
-            )
-
-        # package problem
-        if self.report.get('ProblemType') == 'Package' and \
-            self.report.has_key('Package'):
-
-            title = '[apport] package %s failed to install/upgrade' % \
-                self.report['Package']
-            if self.report.get('ErrorMessage'):
-                title += ': ' + self.report['ErrorMessage'].splitlines()[-1]
-
-            return title
-
-        return None
-
     def open_url(self, url):
         '''Open the given URL in a new browser window.
 
@@ -452,10 +388,14 @@ free memory to automatically analyze the problem and send a report to the develo
                 if message:
                     error += '\n' + message
                 self.ui_error_message(title, error)
-            os.close(r)
+            try:
+                os.close(r)
+            except OSError:
+                pass
             return
 
         os.setsid()
+        os.close(r)
 
         # If we are called through sudo, determine the real user id and run the
         # browser with it to get the user's web browser settings.
@@ -517,12 +457,19 @@ free memory to automatically analyze the problem and send a report to the develo
         '''Upload the current report to the tracking system and guide the user
         to its web page.'''
 
+        global __upload_progress
+        __upload_progress = None
+
+        def progress_callback(sent, total):
+            global __upload_progress
+            __upload_progress = float(sent)/total
+
         self.ui_start_upload_progress()
-        upthread = REThread.REThread(target=upload_launchpad_blob,
-            args=(self.report,))
+        upthread = REThread.REThread(target=self.crashdb.upload,
+            args=(self.report, progress_callback))
         upthread.start()
         while upthread.isAlive():
-            self.ui_set_upload_progress(None)
+            self.ui_set_upload_progress(__upload_progress)
             try:
                 upthread.join(0.1)
             except KeyboardInterrupt:
@@ -530,7 +477,7 @@ free memory to automatically analyze the problem and send a report to the develo
         if upthread.exc_info():
             self.ui_error_message(_('Network problem'),
                 "%s:\n\n%s" % (
-                    _('Could not upload report data to Launchpad'),
+                    _('Could not upload report data to crash database'),
                     str(upthread.exc_info()[1])
                 ))
             return
@@ -538,21 +485,9 @@ free memory to automatically analyze the problem and send a report to the develo
         ticket = upthread.return_value()
         self.ui_stop_upload_progress()
 
-        if ticket:
-            args = {}
-            title = self.create_crash_bug_title()
-            if title:
-                args['field.title'] = title
-
-            if self.report.has_key('SourcePackage'):
-                self.open_url('https://launchpad.net/ubuntu/+source/%s/+filebug/%s?%s' % (
-                    self.report['SourcePackage'], ticket, urllib.urlencode(args)))
-            else:
-                self.open_url('https://launchpad.net/ubuntu/+filebug/%s?%s' % (
-                    ticket, urllib.urlencode(args)))
-        else:
-            self.ui_error_message(_('Launchpad problem'),
-                _('Launchpad did not return a ticket number for the uploaded data.'))
+        url = self.crashdb.get_comment_url(self.report, ticket)
+        if url:
+            self.open_url(url)
 
     def load_report(self, path):
         '''Load report from given path and do some consistency checks.
@@ -563,13 +498,13 @@ free memory to automatically analyze the problem and send a report to the develo
 
         try:
             self.report = apport.Report()
-            self.report.load(open(path))
+            self.report.load(open(path), binary='compressed')
         except MemoryError:
             self.report = None
             self.ui_error_message(_('Memory exhaustion'),
                 _('Your system does not have enough memory to process this crash report.'))
             return False
-        except (TypeError, ValueError):
+        except (TypeError, ValueError, IOError, zlib.error):
             self.report = None
             self.ui_error_message(_('Invalid problem report'),
                 _('This problem report is damaged and cannot be processed.'))
@@ -640,7 +575,7 @@ might be helpful for the developers.'))
         - Valid values for the 'blacklist' key: True or False (True will cause
           the invocation of report.mark_ignore()).'''
 
-        raise Exception, 'this function must be overridden by subclasses'
+        raise NotImplementedError, 'this function must be overridden by subclasses'
 
     def ui_present_package_error(self, desktopentry):
         '''Inform that a package installation/upgrade failure has happened for
@@ -649,7 +584,7 @@ might be helpful for the developers.'))
         Return the action: ignore ('cancel'), or report a bug about the problem
         ('report').'''
 
-        raise Exception, 'this function must be overridden by subclasses'
+        raise NotImplementedError, 'this function must be overridden by subclasses'
 
     def ui_present_kernel_error(self, desktopentry):
         '''Inform that a kernel Oops has happened for self.report and
@@ -658,7 +593,7 @@ might be helpful for the developers.'))
         Return the action: ignore ('cancel'), or report a bug about the problem
         ('report').'''
 
-        raise Exception, 'this function must be overridden by subclasses'
+        raise NotImplementedError, 'this function must be overridden by subclasses'
 
     def ui_present_report_details(self):
         '''Show details of the bug report and choose between sending a complete
@@ -671,23 +606,23 @@ might be helpful for the developers.'))
         Return the action: send full report ('full'), send reduced report
         ('reduced'), or do not send anything ('cancel').'''
 
-        raise Exception, 'this function must be overridden by subclasses'
+        raise NotImplementedError, 'this function must be overridden by subclasses'
 
     def ui_info_message(self, title, text):
         '''Show an information message box with given title and text.'''
 
-        raise Exception, 'this function must be overridden by subclasses'
+        raise NotImplementedError, 'this function must be overridden by subclasses'
 
     def ui_error_message(self, title, text):
         '''Show an error message box with given title and text.'''
 
-        raise Exception, 'this function must be overridden by subclasses'
+        raise NotImplementedError, 'this function must be overridden by subclasses'
 
     def ui_start_info_collection_progress(self):
         '''Open a window with an indefinite progress bar, telling the user to
         wait while debug information is being collected.'''
 
-        raise Exception, 'this function must be overridden by subclasses'
+        raise NotImplementedError, 'this function must be overridden by subclasses'
 
     def ui_pulse_info_collection_progress(self):
         '''Advance the progress bar in the debug data collection progress
@@ -695,18 +630,18 @@ might be helpful for the developers.'))
 
         This function is called every 100 ms.'''
 
-        raise Exception, 'this function must be overridden by subclasses'
+        raise NotImplementedError, 'this function must be overridden by subclasses'
 
     def ui_stop_info_collection_progress(self):
         '''Close debug data collection progress window.'''
 
-        raise Exception, 'this function must be overridden by subclasses'
+        raise NotImplementedError, 'this function must be overridden by subclasses'
 
     def ui_start_upload_progress(self):
         '''Open a window with an definite progress bar, telling the user to
         wait while debug information is being uploaded.'''
 
-        raise Exception, 'this function must be overridden by subclasses'
+        raise NotImplementedError, 'this function must be overridden by subclasses'
 
     def ui_set_upload_progress(self, progress):
         '''Set the progress bar in the debug data upload progress
@@ -715,12 +650,12 @@ might be helpful for the developers.'))
 
         This function is called every 100 ms.'''
 
-        raise Exception, 'this function must be overridden by subclasses'
+        raise NotImplementedError, 'this function must be overridden by subclasses'
 
     def ui_stop_upload_progress(self):
         '''Close debug data upload progress window.'''
 
-        raise Exception, 'this function must be overridden by subclasses'
+        raise NotImplementedError, 'this function must be overridden by subclasses'
 
     def ui_shutdown(self):
         '''This is called right before terminating the program and can be used
@@ -733,17 +668,33 @@ might be helpful for the developers.'))
 #
 
 if  __name__ == '__main__':
-    import unittest, shutil, signal
+    import unittest, shutil, signal, tempfile
 
     class _TestSuiteUserInterface(UserInterface):
         '''Concrete UserInterface suitable for automatic testing.'''
 
         def __init__(self):
+            # use our dummy crashdb
+            self.crashdb_conf = tempfile.NamedTemporaryFile()
+            print >> self.crashdb_conf, '''default = 'testsuite'
+databases = {
+    'testsuite': { 
+        'impl': 'memory',
+        'bug_pattern_base': None
+    }
+}
+'''
+            self.crashdb_conf.flush()
+
+            os.environ['APPORT_CRASHDB_CONF'] = self.crashdb_conf.name
+
             UserInterface.__init__(self)
 
-            # state of info collection progress dialog
+            # state of progress dialogs
             self.ic_progress_active = False
             self.ic_progress_pulses = 0 # count the pulses
+            self.upload_progress_active = False
+            self.upload_progress_pulses = 0
 
             # these store the choices the ui_present_* calls do
             self.present_crash_response = None
@@ -794,17 +745,20 @@ if  __name__ == '__main__':
         def ui_stop_info_collection_progress(self):
             self.ic_progress_active = False
 
+        def ui_start_upload_progress(self):
+            self.upload_progress_pulses = 0
+            self.upload_progress_active = True
+
+        def ui_set_upload_progress(self, progress):
+            assert self.upload_progress_active
+            self.upload_progress_pulses += 1
+
+        def ui_stop_upload_progress(self):
+            self.upload_progress_active = False
+
         def open_url(self, url):
             self.opened_url = url
 
-        def file_report(self):
-            # the report can be accessed as self.ui.report, thus we do not do
-            # anything about sending it anywhere
-            if self.report.has_key('SourcePackage'):
-                self.open_url('https://bug.tracker/%s' % self.report['SourcePackage'])
-            else:
-                self.open_url('https://bug.tracker')
-
     class _UserInterfaceTest(unittest.TestCase):
         def setUp(self):
             # we test a few strings, don't get confused by translations
@@ -988,108 +942,6 @@ CoreDump: base64
             self.assertEqual(self.ui.ic_progress_active, False,
                 'progress dialog for package bug info collection finished')
 
-        def test_create_crash_bug_title(self):
-            '''Test create_crash_bug_title().'''
-
-            self.ui.report = apport.Report()
-            self.assertEqual(self.ui.create_crash_bug_title(), None)
-
-            # named signal crash
-            self.ui.report['Signal'] = '11'
-            self.ui.report['ExecutablePath'] = '/bin/bash'
-            self.ui.report['StacktraceTop'] = '''foo()
-bar(x=3)
-baz()
-'''
-            self.assertEqual(self.ui.create_crash_bug_title(),
-                '[apport] bash crashed with SIGSEGV in foo()')
-
-            # unnamed signal crash
-            self.ui.report['Signal'] = '42'
-            self.assertEqual(self.ui.create_crash_bug_title(),
-                '[apport] bash crashed with signal 42 in foo()')
-
-            # do not crash on empty StacktraceTop
-            self.ui.report['StacktraceTop'] = ''
-            self.assertEqual(self.ui.create_crash_bug_title(),
-                '[apport] bash crashed with signal 42')
-
-            # do not create bug title with unknown function name
-            self.ui.report['StacktraceTop'] = '??()\nfoo()'
-            self.assertEqual(self.ui.create_crash_bug_title(),
-                '[apport] bash crashed with signal 42 in foo()')
-
-            # if we do not know any function name, don't mention ??
-            self.ui.report['StacktraceTop'] = '??()\n??()'
-            self.assertEqual(self.ui.create_crash_bug_title(),
-                '[apport] bash crashed with signal 42')
-
-            # Python crash
-            self.ui.report = apport.Report()
-            self.ui.report['ExecutablePath'] = '/usr/share/apport/apport-gtk'
-            self.ui.report['Traceback'] = '''Traceback (most recent call last):
-  File "/usr/share/apport/apport-gtk", line 202, in <module>
-    app.run_argv()
-  File "/var/lib/python-support/python2.5/apport/ui.py", line 161, in run_argv
-    self.run_crashes()
-  File "/var/lib/python-support/python2.5/apport/ui.py", line 104, in run_crashes
-    self.run_crash(f)
-  File "/var/lib/python-support/python2.5/apport/ui.py", line 115, in run_crash
-    response = self.ui_present_crash(desktop_entry)
-  File "/usr/share/apport/apport-gtk", line 67, in ui_present_crash
-    subprocess.call(['pgrep', '-x',
-NameError: global name 'subprocess' is not defined'''
-            self.assertEqual(self.ui.create_crash_bug_title(),
-                '[apport] apport-gtk crashed with NameError in ui_present_crash()')
-
-            # slightly weird Python crash
-            self.ui.report = apport.Report()
-            self.ui.report['ExecutablePath'] = '/usr/share/apport/apport-gtk'
-            self.ui.report['Traceback'] = '''TypeError: Cannot create a consistent method resolution
-order (MRO) for bases GObject, CanvasGroupableIface, CanvasGroupable'''
-            self.assertEqual(self.ui.create_crash_bug_title(),
-                '[apport] apport-gtk crashed with TypeError: Cannot create a consistent method resolution')
-
-            # package install problem
-            self.ui.report = apport.Report('Package')
-            self.ui.report['Package'] = 'bash'
-
-            # no ErrorMessage
-            self.assertEqual(self.ui.create_crash_bug_title(),
-                '[apport] package bash failed to install/upgrade')
-
-            # empty ErrorMessage
-            self.ui.report['ErrorMessage'] = ''
-            self.assertEqual(self.ui.create_crash_bug_title(),
-                '[apport] package bash failed to install/upgrade')
-
-            # nonempty ErrorMessage
-            self.ui.report['ErrorMessage'] = 'botched\nnot found\n'
-            self.assertEqual(self.ui.create_crash_bug_title(),
-                '[apport] package bash failed to install/upgrade: not found')
-
-            # matching package/system architectures
-            self.ui.report['Signal'] = '11'
-            self.ui.report['ExecutablePath'] = '/bin/bash'
-            self.ui.report['StacktraceTop'] = '''foo()
-bar(x=3)
-baz()
-'''
-            self.ui.report['PackageArchitecture'] = 'amd64'
-            self.ui.report['Architecture'] = 'amd64'
-            self.assertEqual(self.ui.create_crash_bug_title(),
-                '[apport] bash crashed with SIGSEGV in foo()')
-
-            # non-native package (on multiarch)
-            self.ui.report['PackageArchitecture'] = 'i386'
-            self.assertEqual(self.ui.create_crash_bug_title(),
-                '[apport] bash crashed with SIGSEGV in foo() [non-native i386 package]')
-
-            # Arch: all package (matches every system architecture)
-            self.ui.report['PackageArchitecture'] = 'all'
-            self.assertEqual(self.ui.create_crash_bug_title(),
-                '[apport] bash crashed with SIGSEGV in foo()')
-
         def test_handle_duplicate(self):
             '''Test handle_duplicate().'''
 
@@ -1112,7 +964,7 @@ baz()
             self.ui.run_report_bug()
             self.assertEqual(self.ui.msg_severity, None)
             self.assertEqual(self.ui.msg_title, None)
-            self.assertNotEqual(self.ui.opened_url, None)
+            self.assertEqual(self.ui.opened_url, 'http://bug.net/%i' % self.ui.crashdb.latest_id())
 
             self.assert_(set(['Date', 'Uname', 'DistroRelease', 'ProblemType']).issubset(
                 set(self.ui.report.keys())), 'report has required fields')
@@ -1126,13 +978,20 @@ baz()
 
             self.assertEqual(self.ui.msg_severity, None)
             self.assertEqual(self.ui.msg_title, None)
-            self.assertNotEqual(self.ui.opened_url, None)
+            self.assertEqual(self.ui.opened_url, 'http://bash.bug.net/%i' % self.ui.crashdb.latest_id())
 
             self.assert_(self.ui.ic_progress_pulses > 0)
             self.assertEqual(self.ui.report['SourcePackage'], 'bash')
             self.assert_('Dependencies' in self.ui.report.keys())
             self.assertEqual(self.ui.report['ProblemType'], 'Bug')
 
+            # should not crash on nonexisting package
+            sys.argv = ['ui-test', '-f', '-p', 'nonexisting_gibberish']
+            self.ui = _TestSuiteUserInterface()
+            self.ui.run_argv()
+
+            self.assertEqual(self.ui.msg_severity, 'error')
+
         def test_run_report_bug_pid(self):
             '''Test run_report_bug() for a pid.'''
 
@@ -1161,7 +1020,7 @@ baz()
 
             self.assertEqual(self.ui.msg_severity, None)
             self.assertEqual(self.ui.msg_title, None)
-            self.assertNotEqual(self.ui.opened_url, None)
+            self.assertEqual(self.ui.opened_url, 'http://coreutils.bug.net/%i' % self.ui.crashdb.latest_id())
             self.assert_(self.ui.ic_progress_pulses > 0)
 
         def test_run_report_bug_wrong_pid(self):
@@ -1183,6 +1042,19 @@ baz()
             self.ui = _TestSuiteUserInterface()
             self.ui.run_argv()
 
+        def test_run_report_bug_noperm_pid(self):
+            '''Test run_report_bug() for a pid which runs as a different user.'''
+
+            assert os.getuid() > 0, 'this test must not be run as root'
+
+            # silently ignore missing PID; this happens when the user closes
+            # the application prematurely
+            sys.argv = ['ui-test', '-f', '-P', '1']
+            self.ui = _TestSuiteUserInterface()
+            self.ui.run_argv()
+
+            self.assertEqual(self.ui.msg_severity, 'error')
+
         def test_run_crash(self):
             '''Test run_crash().'''
 
@@ -1206,6 +1078,7 @@ baz()
             r = apport.Report()
             r['ExecutablePath'] = test_executable
             r['CoreDump'] = (coredump,)
+            r['Signal'] = '11'
             r.add_proc_info(pid)
             r.add_user_info()
 
@@ -1231,7 +1104,8 @@ baz()
             self.ui.present_crash_response = {'action': 'report', 'blacklist': False }
             self.ui.present_details_response = 'cancel'
             self.ui.run_crash(report_file)
-            self.assertEqual(self.ui.msg_severity, None)
+            self.assertEqual(self.ui.msg_severity, None, 'has %s message: %s: %s' % (
+                self.ui.msg_severity, str(self.ui.msg_title), str(self.ui.msg_text)))
             self.assertEqual(self.ui.msg_title, None)
             self.assertEqual(self.ui.opened_url, None)
             self.assertNotEqual(self.ui.ic_progress_pulses, 0)
@@ -1244,7 +1118,7 @@ baz()
             self.ui.run_crash(report_file)
             self.assertEqual(self.ui.msg_severity, None)
             self.assertEqual(self.ui.msg_title, None)
-            self.assertNotEqual(self.ui.opened_url, None)
+            self.assertEqual(self.ui.opened_url, 'http://coreutils.bug.net/%i' % self.ui.crashdb.latest_id())
             self.assertNotEqual(self.ui.ic_progress_pulses, 0)
 
             self.assert_('SourcePackage' in self.ui.report.keys())
@@ -1252,6 +1126,7 @@ baz()
             self.assert_('Stacktrace' in self.ui.report.keys())
             self.assertEqual(self.ui.report['ProblemType'], 'Crash')
             self.assert_(len(self.ui.report['CoreDump']) > 10000)
+            self.assert_(self.ui.report['Title'].startswith('cat crashed with SIGSEGV'))
 
             # report in crash notification dialog, send reduced report
             r.write(open(report_file, 'w'))
@@ -1261,7 +1136,7 @@ baz()
             self.ui.run_crash(report_file)
             self.assertEqual(self.ui.msg_severity, None)
             self.assertEqual(self.ui.msg_title, None)
-            self.assertNotEqual(self.ui.opened_url, None)
+            self.assertEqual(self.ui.opened_url, 'http://coreutils.bug.net/%i' % self.ui.crashdb.latest_id())
             self.assertNotEqual(self.ui.ic_progress_pulses, 0)
 
             self.assert_('SourcePackage' in self.ui.report.keys())
@@ -1285,6 +1160,28 @@ baz()
 
             self.assert_(self.ui.report.check_ignored())
 
+        def test_run_crash_argv_file(self):
+            '''Test run_crash() through a file specified on the command line.'''
+
+            self.report['Package'] = 'bash'
+            self.report['UnsupportableReason'] = 'It stinks.'
+            self.update_report_file()
+
+            sys.argv = ['ui-test', '-c', self.report_file.name]
+            self.ui = _TestSuiteUserInterface()
+            self.ui.run_argv()
+
+            self.assert_('It stinks.' in self.ui.msg_text, '%s: %s' %
+                (self.ui.msg_title, self.ui.msg_text))
+            self.assertEqual(self.ui.msg_severity, 'info')
+
+            # should not die with an exception on an invalid name
+            sys.argv = ['ui-test', '-c', '/nonexisting.crash' ]
+            self.ui = _TestSuiteUserInterface()
+            self.ui.run_argv()
+            self.assertEqual(self.ui.msg_severity, 'error')
+            self.assert_('/nonexisting.crash' in self.ui.msg_text, self.ui.msg_text)
+
         def test_run_crash_unsupportable(self):
             '''Test run_crash() on a crash with the UnsupportableReason
             field.'''
@@ -1371,8 +1268,8 @@ baz()
 
             # generate crash report
             r = apport.Report('Package')
-            r['Package'] = 'libfoo1'
-            r['SourcePackage'] = 'foo'
+            r['Package'] = 'bash'
+            r['SourcePackage'] = 'bash'
             r['ErrorMessage'] = 'It broke'
             r['VarLogPackagerlog'] = 'foo\nbar'
             r.add_os_info()
@@ -1397,12 +1294,17 @@ baz()
             self.ui.run_crash(report_file)
             self.assertEqual(self.ui.msg_severity, None)
             self.assertEqual(self.ui.msg_title, None)
-            self.assertNotEqual(self.ui.opened_url, None)
+            self.assertEqual(self.ui.opened_url, 'http://bash.bug.net/%i' % self.ui.crashdb.latest_id())
 
             self.assert_('SourcePackage' in self.ui.report.keys())
             self.assert_('Package' in self.ui.report.keys())
             self.assertEqual(self.ui.report['ProblemType'], 'Package')
 
+            # verify that additional information has been collected
+            self.assert_('PackageArchitecture' in self.ui.report.keys())
+            self.assert_('DistroRelease' in self.ui.report.keys())
+            self.assert_('Uname' in self.ui.report.keys())
+
         def test_run_crash_kernel(self):
             '''Test run_crash() for a kernel error.'''
 
@@ -1433,7 +1335,7 @@ baz()
             self.ui.run_crash(report_file)
             self.assertEqual(self.ui.msg_severity, None)
             self.assertEqual(self.ui.msg_title, None)
-            self.assertNotEqual(self.ui.opened_url, None)
+            self.assertEqual(self.ui.opened_url, 'http://linux-source-2.6.20.bug.net/%i' % self.ui.crashdb.latest_id())
 
             self.assert_('SourcePackage' in self.ui.report.keys())
             self.assertEqual(self.ui.report['ProblemType'], 'Kernel')
