@@ -12,12 +12,56 @@ option) any later version.  See http://www.gnu.org/copyleft/gpl.html for
 the full text of the license.
 '''
 
-import bz2, zlib, base64, time, UserDict, sys, gzip
+import zlib, base64, time, UserDict, sys, gzip, struct
 from cStringIO import StringIO
+from email.Encoders import encode_base64
 from email.MIMEMultipart import MIMEMultipart
 from email.MIMEBase import MIMEBase
 from email.MIMEText import MIMEText
 
+class CompressedValue:
+    '''Represent a ProblemReport value which is gzip compressed.'''
+
+    def __init__(self, value=None, name=None):
+        '''Initialize an empty CompressedValue object with an optional name.'''
+
+        self.gzipvalue = None
+        self.name = name
+        if value:
+            self.set_value(value)
+
+    def set_value(self, value):
+        '''Set uncompressed value.'''
+
+        out = StringIO()
+        gzip.GzipFile(self.name, mode='wb', fileobj=out).write(value)
+        self.gzipvalue = out.getvalue()
+
+    def get_value(self):
+        '''Return uncompressed value.'''
+
+        if not self.gzipvalue:
+            return None
+
+        return gzip.GzipFile(fileobj=StringIO(self.gzipvalue)).read()
+
+    def write(self, file):
+        '''Write uncompressed value into given file-like object.'''
+
+        assert self.gzipvalue
+        gz = gzip.GzipFile(fileobj=StringIO(self.gzipvalue))
+        while True:
+            block = gz.read(1048576)
+            if not block:
+                break
+            file.write(block)
+
+    def __len__(self):
+        '''Return length of uncompressed value.'''
+
+        assert self.gzipvalue
+        return int(struct.unpack("<L", self.gzipvalue[-4:])[0])
+
 class ProblemReport(UserDict.IterableUserDict):
     def __init__(self, type = 'Crash', date = None):
         '''Initialize a fresh problem report.
@@ -37,7 +81,12 @@ class ProblemReport(UserDict.IterableUserDict):
         control file format.
 
         if binary is False, binary data is not loaded; the dictionary key is
-        created, but its value will be an empty string.'''
+        created, but its value will be an empty string. If it is true, it is
+        transparently uncompressed and available as dictionary string values.
+        If binary is 'compressed', the compressed value is retained, and the
+        dictionary value will be a CompressedValue object. This is useful if
+        the compressed value is still useful (to avoid recompression if the
+        file needs to be written back).'''
 
         self.data.clear()
         key = None
@@ -55,24 +104,27 @@ class ProblemReport(UserDict.IterableUserDict):
                     if bd:
                         value += bd.decompress(l)
                     else:
-                        # lazy initialization of bd; fall back to bzip2 if gzip
-                        # fails
-                        bd = zlib.decompressobj()
-                        try:
-                            value += bd.decompress(l)
-                        except zlib.error:
-                            bd = bz2.BZ2Decompressor()
-                            value += bd.decompress(l)
+                        if binary == 'compressed':
+                            value.gzipvalue += l
+                        else:
+                            # lazy initialization of bd
+                            # skip gzip header, if present
+                            if l.startswith('\037\213\010'): 
+                                bd = zlib.decompressobj(-zlib.MAX_WBITS)
+                                value = ''
+                            else:
+                                # legacy zlib-only format used default block
+                                # size
+                                bd = zlib.decompressobj()
+                                value += bd.decompress(l)
                 else:
                     if len(value) > 0:
                         value += '\n'
                     value += line[1:-1]
             else:
                 if b64_block:
-                    try:
+                    if bd:
                         value += bd.flush()
-                    except AttributeError:
-                        pass # bz2 decompressor has no flush()
                     b64_block = False
                     bd = None
                 if key:
@@ -81,7 +133,11 @@ class ProblemReport(UserDict.IterableUserDict):
                 (key, value) = line.split(':', 1)
                 value = value.strip()
                 if value == 'base64':
-                    value = ''
+                    if binary == 'compressed':
+                        value = CompressedValue(key)
+                        value.gzipvalue = ''
+                    else:
+                        value = ''
                     b64_block = True
 
         if key != None:
@@ -111,12 +167,13 @@ class ProblemReport(UserDict.IterableUserDict):
         load() are written (i. e. those returned by new_keys()).
 
         If a value is a string, it is written directly. Otherwise it must be a
-        tuple of the form (file, encode=True, limit=None).
+        tuple of the form (file, encode=True, limit=None, fail_on_empty=False).
         The first argument can be a file name or a file-like object,
         which will be read and its content will become the value of this key.
         'encode' specifies whether the contents will be
-        zlib compressed and base64-encoded (this defaults to True). If limit is
-        set to a positive integer, the entire key will be removed.
+        gzip compressed and base64-encoded (this defaults to True). If limit is
+        set to a positive integer, the entire key will be removed. If
+        fail_on_empty is True, reading zero bytes will cause an IOError.
         '''
 
         # sort keys into ASCII non-ASCII/binary attachment ones, so that
@@ -155,11 +212,16 @@ class ProblemReport(UserDict.IterableUserDict):
                 else:
                     limit = None
 
+                fail_on_empty = len(v) >= 4 and v[3]
+
                 if hasattr(v[0], 'read'):
                     v = v[0].read() # file-like object
                 else:
                     v = open(v[0]).read() # file name
 
+                if fail_on_empty and len(v) == 0:
+                    raise IOError, 'did not get any data for field ' + k
+
                 if limit != None and len(v) > limit:
                     del self.data[k]
                     continue
@@ -172,7 +234,7 @@ class ProblemReport(UserDict.IterableUserDict):
                 # single line value
                 print >> file, k + ':', v
 
-        # now write the binary keys with zlib compression and base64 encoding
+        # now write the binary keys with gzip compression and base64 encoding
         for k in binkeys:
             v = self.data[k]
             limit = None
@@ -180,9 +242,19 @@ class ProblemReport(UserDict.IterableUserDict):
 
             curr_pos = file.tell()
             file.write (k + ': base64\n ')
-            bc = zlib.compressobj()
+
+            # write gzip header
+            gzip_header = '\037\213\010\010\000\000\000\000\002\377' + k + '\000'
+            file.write(base64.b64encode(gzip_header))
+            file.write('\n ')
+            crc = zlib.crc32('')
+
+            bc = zlib.compressobj(9, zlib.DEFLATED, -zlib.MAX_WBITS,
+                zlib.DEF_MEM_LEVEL, 0)
             # direct value
             if hasattr(v, 'find'):
+                size += len(v)
+                crc = zlib.crc32(v, crc)
                 outblock = bc.compress(v)
                 if outblock:
                     file.write(base64.b64encode(outblock))
@@ -198,13 +270,15 @@ class ProblemReport(UserDict.IterableUserDict):
                     f = open(v[0]) # file name
                 while True:
                     block = f.read(1048576)
+                    size += len(block)
+                    crc = zlib.crc32(block, crc)
                     if limit != None:
-                        size += len(block)
                         if size > limit:
                             # roll back
                             file.seek(curr_pos)
                             file.truncate(curr_pos)
                             del self.data[k]
+                            crc = None
                             break
                     if block:
                         outblock = bc.compress(block)
@@ -214,9 +288,19 @@ class ProblemReport(UserDict.IterableUserDict):
                     else:
                         break
 
+                if len(v) >= 4 and v[3]:
+                    if size == 0:
+                        raise IOError, 'did not get any data for field %s from %s' % (k, str(v[0]))
+
             # flush compressor and write the rest
             if not limit or size <= limit:
-                file.write(base64.b64encode(bc.flush()))
+                block = bc.flush()
+                # append gzip trailer: crc (32 bit) and size (32 bit)
+                if crc:
+                    block += struct.pack("<L", crc & 0xFFFFFFFFL)
+                    block += struct.pack("<L", size & 0xFFFFFFFFL)
+
+                file.write(base64.b64encode(block))
                 file.write('\n')
 
     def add_to_existing(self, reportfile, keep_times=False):
@@ -238,21 +322,22 @@ class ProblemReport(UserDict.IterableUserDict):
                 os.utime(reportfile, (st.st_atime, st.st_mtime))
             os.chmod(reportfile, st.st_mode)
 
-    def write_mime(self, file, attach_treshold = 5, preamble=None):
+    def write_mime(self, file, attach_treshold = 5, extra_headers={}):
         '''Write information into the given file-like object, using
         MIME/Multipart RFC 2822 format (i. e. an email with attachments).
 
-        If a value is a string, it is written directly. Otherwise it must be a
-        tuple containing the source file and an optional boolean value (in that
-        order); the first argument can be a file name or a file-like object,
-        which will be read and its content will become the value of this key.
-        The file will be gzip compressed, unless the key already ends in .gz.
+        If a value is a string or a CompressedValue, it is written directly.
+        Otherwise it must be a tuple containing the source file and an optional
+        boolean value (in that order); the first argument can be a file name or
+        a file-like object, which will be read and its content will become the
+        value of this key.  The file will be gzip compressed, unless the key
+        already ends in .gz.
 
         attach_treshold specifies the maximum number of lines for a value to be
         included into the first inline text part. All bigger values (as well as
         all non-ASCII ones) will become an attachment.
 
-        A MIME preamble can be specified, too.
+        Extra MIME preamble headers can be specified, too, as a dictionary.
         '''
 
         keys = self.data.keys()
@@ -269,19 +354,23 @@ class ProblemReport(UserDict.IterableUserDict):
             v = self.data[k]
             attach_value = None
 
+            # compressed values are ready for attaching in gzip form
+            if isinstance(v, CompressedValue):
+                attach_value = v.gzipvalue
+
             # if it's a tuple, we have a file reference; read the contents
             # and gzip it
-            if not hasattr(v, 'find'):
+            elif not hasattr(v, 'find'):
                 attach_value = ''
                 if hasattr(v[0], 'read'):
                     f = v[0] # file-like object
                 else:
                     f = open(v[0]) # file name
                 if k.endswith('.gz'):
-                    attach_value = StringIO(f.read())
+                    attach_value = f.read()
                 else:
-                    attach_value = StringIO()
-                    gf = gzip.GzipFile(k, mode='wb', fileobj=attach_value)
+                    io = StringIO()
+                    gf = gzip.GzipFile(k, mode='wb', fileobj=io)
                     while True:
                         block = f.read(1048576)
                         if block:
@@ -289,17 +378,15 @@ class ProblemReport(UserDict.IterableUserDict):
                         else:
                             gf.close()
                             break
+                    attach_value = io.getvalue()
                 f.close()
 
             # binary value
             elif self._is_binary(v):
                 if k.endswith('.gz'):
-                    attach_value = StringIO(v)
+                    attach_value = v
                 else:
-                    attach_value = StringIO()
-                    gf = gzip.GzipFile(k, mode='wb', fileobj=attach_value)
-                    gf.write(v)
-                    gf.close()
+                    attach_value = CompressedValue(v, k).gzipvalue
 
             # if we have an attachment value, create an attachment
             if attach_value:
@@ -308,7 +395,8 @@ class ProblemReport(UserDict.IterableUserDict):
                     att.add_header('Content-Disposition', 'attachment', filename=k)
                 else:
                     att.add_header('Content-Disposition', 'attachment', filename=k+'.gz')
-                att.set_payload(attach_value.getvalue())
+                att.set_payload(attach_value)
+                encode_base64(att)
                 attachments.append(att)
             else:
                 # plain text value
@@ -332,8 +420,8 @@ class ProblemReport(UserDict.IterableUserDict):
         attachments.insert(0, att)
 
         msg = MIMEMultipart()
-        if preamble:
-            msg.preamble = preamble
+        for k, v in extra_headers.iteritems():
+            msg.add_header(k, v)
         for a in attachments:
             msg.attach(a)
 
@@ -342,8 +430,9 @@ class ProblemReport(UserDict.IterableUserDict):
     def __setitem__(self, k, v):
         assert hasattr(k, 'isalnum')
         assert k.replace('.', '').isalnum()
-        # value must be a string or a file reference (tuple (string|file [, bool]))
-        assert (hasattr(v, 'isalnum') or
+        # value must be a string or a CompressedValue or a file reference
+        # (tuple (string|file [, bool]))
+        assert (isinstance(v, CompressedValue) or hasattr(v, 'isalnum') or
             (hasattr(v, '__getitem__') and (
             len(v) == 1 or (len(v) >= 2 and v[1] in (True, False)))
             and (hasattr(v[0], 'isalnum') or hasattr(v[0], 'read'))))
@@ -394,6 +483,33 @@ class _ProblemReportTest(unittest.TestCase):
         self.assertRaises(AssertionError, pr.__setitem__, 'a', ('/tmp/nonexistant', ''))
         self.assertRaises(KeyError, pr.__getitem__, 'Nonexistant')
 
+    def test_compressed_values(self):
+        '''Test handling of CompressedValue values.'''
+
+        large_val = 'A' * 5000000
+
+        pr = ProblemReport()
+        pr['Foo'] = CompressedValue('FooFoo!')
+        pr['Bin'] = CompressedValue()
+        pr['Bin'].set_value('AB' * 10 + '\0' * 10 + 'Z')
+        pr['Large'] = CompressedValue(large_val)
+
+        self.assert_(isinstance(pr['Foo'], CompressedValue))
+        self.assert_(isinstance(pr['Bin'], CompressedValue))
+        self.assertEqual(pr['Foo'].get_value(), 'FooFoo!')
+        self.assertEqual(pr['Bin'].get_value(), 'AB' * 10 + '\0' * 10 + 'Z')
+        self.assertEqual(pr['Large'].get_value(), large_val)
+        self.assertEqual(len(pr['Foo']), 7)
+        self.assertEqual(len(pr['Bin']), 31)
+        self.assertEqual(len(pr['Large']), len(large_val))
+
+        io = StringIO()
+        pr['Bin'].write(io)
+        self.assertEqual(io.getvalue(), 'AB' * 10 + '\0' * 10 + 'Z')
+        io = StringIO()
+        pr['Large'].write(io)
+        self.assertEqual(io.getvalue(), large_val)
+
     def test_write(self):
         '''Test write() and proper formatting.'''
 
@@ -452,14 +568,13 @@ Extra: appended
         pr['Extra'] = 'appended'
         pr.write(io)
 
-        self.assertEqual(io.getvalue(),
-'''ProblemType: Crash
-Date: now!
-File: base64
- eJw=
- c3RyxIAMcBAFAG55BXk=
-Extra: appended
-''')
+        io.seek(0)
+        pr = ProblemReport()
+        pr.load(io)
+
+        self.assertEqual(pr['Date'], 'now!')
+        self.assertEqual(pr['File'], 'AB' * 10 + '\0' * 10 + 'Z')
+        self.assertEqual(pr['Extra'], 'appended')
 
     def test_load(self):
         '''Test load() with various formatting.'''
@@ -550,11 +665,11 @@ Last: foo
 '''ProblemType: Crash
 Date: now!
 Afile: base64
- eJw=
- c3RyxIAMcBAFAG55BXk=
+ H4sICAAAAAAC/0FmaWxlAA==
+ c3RyxIAMcBAFAK/2p9MfAAAA
 File: base64
- eJw=
- c3RyxIAMcBAFAG55BXk=
+ H4sICAAAAAAC/0ZpbGUA
+ c3RyxIAMcBAFAK/2p9MfAAAA
 ''')
 
         # force compression/encoding bool
@@ -580,8 +695,8 @@ File: foo\0bar
 '''ProblemType: Crash
 Date: now!
 File: base64
- eJw=
- S8vPZ0hKLAIACfACeg==
+ H4sICAAAAAAC/0ZpbGUA
+ S8vPZ0hKLAIACq50HgcAAAA=
 ''')
         temp.close()
 
@@ -596,15 +711,59 @@ File: base64
         pr['AscFile'] = (tempasc, False)
         io = StringIO()
         pr.write(io)
+        io.seek(0)
 
-        self.assertEqual(io.getvalue(),
-'''ProblemType: Crash
-AscFile: Hello World
-Date: now!
-BinFile: base64
- eJw=
- c3RyxIAMcBAFAG55BXk=
-''')
+        pr = ProblemReport()
+        pr.load(io)
+        self.assertEqual(pr['BinFile'], tempbin.getvalue())
+        self.assertEqual(pr['AscFile'], tempasc.getvalue())
+
+    def test_write_empty_fileobj(self):
+        '''Test writing a report with a pointer to a file-like object with enforcing non-emptyness.'''
+
+        tempbin = StringIO('')
+        tempasc = StringIO('')
+
+        pr = ProblemReport(date = 'now!')
+        pr['BinFile'] = (tempbin, True, None, True)
+        io = StringIO()
+        self.assertRaises(IOError, pr.write, io)
+
+        pr = ProblemReport(date = 'now!')
+        pr['AscFile'] = (tempasc, False, None, True)
+        io = StringIO()
+        self.assertRaises(IOError, pr.write, io)
+
+    def test_write_delayed_fileobj(self):
+        '''Test writing a report with file pointers and delayed data.'''
+
+        (fout, fin) = os.pipe()
+
+        if os.fork() == 0:
+            os.close(fout)
+            time.sleep(0.3)
+            os.write(fin, 'ab' * 512*1024)
+            time.sleep(0.3)
+            os.write(fin, 'hello')
+            time.sleep(0.3)
+            os.write(fin, ' world')
+            os.close(fin)
+            os._exit(0)
+
+        os.close(fin)
+
+        pr = ProblemReport(date = 'now!')
+        pr['BinFile'] = (os.fdopen(fout),)
+        io = StringIO()
+        pr.write(io)
+        assert os.wait()[1] == 0
+
+        io.seek(0)
+
+        pr2 = ProblemReport()
+        pr2.load(io)
+        self.assert_(pr2['BinFile'].endswith('abhello world'))
+        self.assertEqual(len(pr2['BinFile']), 1048576 + len('hello world'))
 
     def test_read_file(self):
         '''Test reading a report with binary data.'''
@@ -612,8 +771,8 @@ BinFile: base64
         bin_report = '''ProblemType: Crash
 Date: now!
 File: base64
- eJw=
- c3RyxIAMcBAFAG55BXk=
+ H4sICAAAAAAC/0ZpbGUA
+ c3RyxIAMcBAFAK/2p9MfAAAA
 Foo: Bar
 '''
 
@@ -628,13 +787,23 @@ Foo: Bar
         self.assertEqual(pr['File'], '')
         self.assertEqual(pr.has_removed_fields(), True)
 
-    def test_read_file_bzip2(self):
-        '''Test reading a report with binary data (legacy bzip2 compression).'''
+        # test with keeping compressed binary data
+        pr.load(StringIO(bin_report), binary='compressed')
+        self.assertEqual(pr['Foo'], 'Bar')
+        self.assertEqual(pr.has_removed_fields(), False)
+        self.assert_(isinstance(pr['File'], CompressedValue))
+
+        self.assertEqual(pr['File'].get_value(), 'AB' * 10 + '\0' * 10 + 'Z')
+
+    def test_read_file_legacy(self):
+        '''Test reading a report with binary data in legacy format without gzip
+        header.'''
 
         bin_report = '''ProblemType: Crash
 Date: now!
 File: base64
- QlpoOTFBWSZTWc5ays4AAAdGAEEAMAAAECAAMM0AkR6fQsBSDhdyRThQkM5ays4=
+ eJw=
+ c3RyxIAMcBAFAG55BXk=
 Foo: Bar
 '''
 
@@ -681,6 +850,12 @@ Foo: Bar
         pr.write(io2)
         self.assert_(io.getvalue() == io2.getvalue())
 
+        # check gzip compatibility
+        io.seek(0)
+        pr = ProblemReport()
+        pr.load(io, binary='compressed')
+        self.assertEqual(pr['File'].get_value(), data)
+
     def test_size_limit(self):
         '''Test writing and a big random file with a size limit key.'''
 
@@ -703,6 +878,8 @@ Foo: Bar
         pr.write(io)
         temp.close()
 
+        open('/tmp/r', 'w').write(io.getvalue())
+
         # read it again
         io.seek(0)
         pr = ProblemReport()
@@ -741,8 +918,8 @@ Long:
  yyy
 Short: Bar
 File: base64
- eJw=
- c3RyxIAMcBAFAG55BXk=
+ H4sICAAAAAAC/0ZpbGUA
+ c3RyxIAMcBAFAK/2p9MfAAAA
 '''
 
         pr = ProblemReport()
@@ -767,8 +944,8 @@ Short:
  aaa
  bbb
 File: base64
- eJw=
- c3RyxIAMcBAFAG55BXk=
+ H4sICAAAAAAC/0ZpbGUA
+ c3RyxIAMcBAFAK/2p9MfAAAA
 ''')
 
     def test_add_to_existing(self):
@@ -911,6 +1088,7 @@ line♥5!!
         pr['File1.gz'] = (tempgz.name,)
         pr['Value1'] = bin_value
         pr['Value1.gz'] = open(tempgz.name).read()
+        pr['ZValue'] = CompressedValue(bin_value)
         io = StringIO()
         pr.write_mime(io)
         io.seek(0)
@@ -974,21 +1152,34 @@ line♥5!!
         f.seek(0)
         self.assertEqual(gzip.GzipFile(mode='rb', fileobj=f).read(), bin_value)
 
+        # seventh part should be the ZValue: value as gzip'ed attachment;
+        # write_mime should not compress it again
+        part = msg_iter.next()
+        self.assert_(not part.is_multipart())
+        self.assertEqual(part.get_content_type(), 'application/x-gzip')
+        self.assertEqual(part.get_filename(), 'ZValue.gz')
+        f = tempfile.TemporaryFile()
+        f.write(part.get_payload(decode=True))
+        f.seek(0)
+        self.assertEqual(gzip.GzipFile(mode='rb', fileobj=f).read(), bin_value)
+
         # no more parts
         self.assertRaises(StopIteration, msg_iter.next)
 
-    def test_write_mime_preamble(self):
-        '''Test write_mime() with a preamble.'''
+    def test_write_mime_extra_headers(self):
+        '''Test write_mime() with extra headers.'''
 
         pr = ProblemReport(date = 'now!')
         pr['Simple'] = 'bar'
         pr['TwoLine'] = 'first\nsecond\n'
         io = StringIO()
-        pr.write_mime(io, preamble='hello world')
+        pr.write_mime(io, extra_headers={'Greeting': 'hello world', 
+            'Foo': 'Bar'})
         io.seek(0)
 
         msg = email.message_from_file(io)
-        self.assertEqual(msg.preamble, 'hello world')
+        self.assertEqual(msg['Greeting'], 'hello world')
+        self.assertEqual(msg['Foo'], 'Bar')
         msg_iter = msg.walk()
 
         # first part is the multipart container
